from dataclasses import dataclass, field
from pathlib import Path

import click
from tttp.prompter import Prompter

from ttt.config import arg2dict, encoding


@dataclass
class Chunker:
    text: str
    token_chunks: list = field(default_factory=list)
    chunks: list = field(default_factory=list)
    processed_chunks: list = field(default_factory=list)
    params: dict = field(default_factory=dict)
    template_size: int = 0

    def __post_init__(self):
        if self.params.get("template_file", None):
            file = Prompter.find_file(self.params["template_file"])
            self.template_size = len(encoding.encode(Path(file).read_text()))
            self.prompter = Prompter(file)

        self.summary_size: int = self.params.get("summary_size", 500)
        self.chunk_size = self.params.get("chunk_size", self.params.get("token_limit") - self.summary_size)

        self.tokens = encoding.encode(self.text)
        self.tokens_size = len(self.tokens)
        self.in_size = min(self.chunk_size, self.tokens_size) + self.template_size
        self.total_size = self.in_size + self.summary_size

    def needs_chunking(self):
        if self.total_size > self.params.get("token_limit") or self.tokens_size > self.chunk_size:
            click.echo(
                "Prompt is too long. "
                f"Prompt/template/summary: {self.tokens_size}/{self.template_size}/{self.summary_size} "
                f"Token limit tokens: {self.params.get('token_limit')}",
                err=True,
            )
            return True
        return False

    def chunk(self):
        prompt_args = arg2dict(self.params["template_args"])
        self.token_chunks = [self.tokens[i : i + self.chunk_size] for i in range(0, len(self.tokens), self.chunk_size)]
        self.chunks = [encoding.decode(chunk) for chunk in self.token_chunks]
        self.chunks = [self.prompter.prompt(chunk, prompt_args) for chunk in self.chunks]
        return self.chunks



import datetime
import logging
from pathlib import Path

import tiktoken
import tttp
import yaml

encoding = tiktoken.get_encoding("gpt2")

file_path = Path("/tmp/ttt/")
file_path.mkdir(parents=True, exist_ok=True)

log_path = file_path / "logs"
log_path.mkdir(parents=True, exist_ok=True)

timestamp = datetime.datetime.now().replace(microsecond=0).isoformat()
logfile = log_path / f"{timestamp}.log"

logging.basicConfig(
    filename=logfile,
    format="%(asctime)s - %(levelname)s - %(message)s",
    datefmt="%Y-%m-%d %H:%M:%S",
    level=logging.INFO,
)
logger = logging.getLogger()

config_dir = Path().home() / ".config/ttt"
config_path = config_dir / "config.yaml"


def load_config():
    global config
    if not config_path.exists():
        config = {}
        return
    config = yaml.load(config_path.read_text(), Loader=yaml.Loader)


def create_config():
    config_dir.mkdir(parents=True, exist_ok=True)
    config_path.write_text(yaml.dump(TURBO_TEXT_TRANSFORMER_DEFAULT_PARAMS))

    # Find the templates, and make sure they are in the right place
    tttp_dir = Path(tttp.__file__).parent
    new_templates = tttp_dir.parent / "templates"
    templates = config_dir / "templates"
    templates.mkdir(parents=True, exist_ok=True)
    for template in new_templates.glob("*.j2"):
        if not (templates / template.name).exists():
            (templates / template.name).write_text(template.read_text())


TURBO_TEXT_TRANSFORMER_DEFAULT_PARAMS = {"format": "clean", "echo_prompt": False, "backup_path": "/tmp/ttt/"}

OPENAI_DEFAULT_PARAMS = {
    "frequency_penalty": 0,
    "logprobs": 1,
    "max_tokens": 200,
    "model": "gpt-3.5-turbo",
    "n": 1,
    "presence_penalty": 0,
    "stop": None,
    "temperature": 0.9,
    "top_p": 1,
}


load_config()


def arg2dict(args):
    d = {}
    if "=" not in args:
        return d
    for arg in args.split(","):
        k, v = arg.split("=")
        d[k] = v
    return d



#!/usr/bin/env python3

from pathlib import Path

import click

from ttt.chunker import Chunker
from ttt.config import arg2dict, config_path, create_config, load_config
from ttt.models import BaseModel, OpenAIModel


def check_config(reinit):
    """Check that the config file exists."""
    if not reinit and config_path.exists():
        return

    click.echo("Config file not found. Creating one for you...", err=True, color="red")
    create_config()
    openai_api_key = click.prompt("OpenAI API Key", type=str)
    if openai_api_key:
        OpenAIModel.create_config(openai_api_key)

    # Same for other providers...

    load_config()


def prepare_engine_params(params, format):
    """Prepare options for the OpenAI API."""
    params = {k: v for k, v in params.items() if v is not None}

    params["token_limit"] = 4000 if params["model"] in OpenAIModel().large_models else 2048
    if "number" in params:
        params["n"] = params.pop("number")
    if "logprobs" in params and params["logprobs"] == 0:
        if format == "logprobs":
            # If format is logprobs, we need to get the logprobs
            params["logprobs"] = 1
        else:
            params["logprobs"] = None
    return params


def get_prompt(prompt, prompt_file, params):
    """Get the prompt from stdin if it's not provided."""
    if prompt_file:
        prompt = Path(prompt_file).read_text().strip()
        return prompt
    if not prompt:
        click.echo("Reading from stdin... (Ctrl-D to end)", err=True)
        prompt = click.get_text_stream("stdin").read().strip()
        click.echo("Generating...", err=True)
        if not prompt:
            click.echo("No prompt provided. Use the -p flag or pipe a prompt to stdin.", err=True, color="red")
            raise click.Abort()
        params["force"] = True
    return prompt


def chunk(prompt, params):
    prompt_args = arg2dict(params["template_args"])
    chunker = Chunker(prompt, params=params)
    if chunker.needs_chunking():
        if not params["force"]:
            click.confirm("Do you want to chunk the prompt?", abort=True, err=True)
        click.echo("Chunking...", err=True)
        return chunker.chunk()
    if chunker.template_size:
        return [chunker.prompter.prompt(prompt, prompt_args)]
    return [prompt]


@click.command()
@click.argument("prompt", required=False)
@click.option(
    "--format",
    "-f",
    help="Output Format",
    default="clean",
    type=click.Choice(["clean", "json", "logprobs"]),
    show_default=True,
)
@click.option("--reinit", "-R", help="Recreate the config files", is_flag=True, default=False)
@click.option("--echo_prompt", "-e", help="Echo the pormpt in the output", is_flag=True, default=False)
@click.option("--list_models", "-l", help="List available models.", is_flag=True, default=False)
@click.option("--prompt_file", "-P", help="File to load for the prompt", default=None)
@click.option("--template_file", "-t", help="Template to apply to prompt.", default=None, type=str)
@click.option("--template_args", "-x", help="Extra values for the template.", default="")
@click.option("--chunk_size", "-c", help="Max size of chunks", default=None, type=int)
@click.option("--summary_size", "-s", help="Size of chunk summaries", default=None, type=int)
@click.option("--model", "-m", help="Name of the model to use.", default="gpt-3.5-turbo")
@click.option("--number", "-N,n", help="Number of completions.", default=None, type=int)
@click.option("--logprobs", "-L", help="Show logprobs for completion", default=None, type=int)
@click.option("--max_tokens", "-M", help="Max number of tokens to return", default=None, type=int)
@click.option(
    "--temperature", "-T", help="Temperature, [0, 2]-- 0 is deterministic, >0.9 is creative.", default=None, type=int
)
@click.option("--force", "-F", help="Force chunking of prompt", is_flag=True, default=False)
def main(prompt, format, reinit, echo_prompt, list_models, prompt_file, **params):
    # click.echo(params, err=True)
    check_config(reinit)
    params = prepare_engine_params(params, format)
    options = {"params": params, "format": format, "echo_prompt": echo_prompt}

    sink = click.get_text_stream("stdout")
    oam = OpenAIModel(**options)
    if list_models:
        sink.write("\n".join(oam.list))
        return

    prompt = get_prompt(prompt, prompt_file, params)
    prompts = chunk(prompt, params)
    if params["model"] in oam.list:
        responses = [oam.gen(prompt) for prompt in prompts]
        sink.write("\n".join(responses))
        return

    if params["model"] == "test":
        bm = BaseModel(**options)
        responses = [bm.gen(prompt) for prompt in prompts]
        sink.write("\n".join(responses))
        return

    click.echo("Model not found. Use the -l flag to list available models.", err=True, color="red")



import datetime
import json
import math
from dataclasses import dataclass, field
from pathlib import Path
from typing import Optional

import click
import openai
import yaml
from colored import attr, bg, fg

from ttt.config import config, config_dir, encoding


@dataclass
class ProbColors:
    # Colors of foreground and background for different probabilities
    prob_1 = (239, 49)
    prob_2 = (239, 78)
    prob_3 = (195, 145)
    prob_4 = (195, 173)
    prob_5 = (195, 209)
    prob_6 = (195, 203)

    @staticmethod
    def choose_color(logprob):
        prob = math.exp(logprob)
        if prob >= 0.8:
            return ProbColors.prob_1
        if prob >= 0.6:
            return ProbColors.prob_2
        if prob >= 0.4:
            return ProbColors.prob_3
        if prob >= 0.2:
            return ProbColors.prob_4
        if prob >= 0.05:
            return ProbColors.prob_5
        return ProbColors.prob_6


@dataclass
class Formatter:
    operator: str
    format: str = config.get("format", "clean")
    echo_prompt: bool = config.get("echo_prompt", False)

    def format_response(self, response):
        if self.operator == "OpenAI":
            return self._openai(response)
        return self._base(response)

    def _openai(self, response):
        if self.format == "json":
            response = self._clean_json(response)
            return json.dumps(response, indent=4)
        if self.format == "logprobs":
            response = self._logprobs(response)
        if self.echo_prompt:
            return "\n".join([response["params"]["prompt"] + c["text"] for c in response["choices"]])
        return "\n".join([c["text"] for c in response["choices"]])

    def _clean_json(self, response):
        for choice in response["choices"]:
            if "logprobs" not in choice:
                continue
            if choice.get("logprobs", None) is None:
                del choice["logprobs"]
                del choice["index"]
                continue

            choice["token_logprobs"] = choice.get("logprobs", {}).get("token_logprobs", None)
            choice["logprob_offset"] = choice.get("logprobs", {}).get("text_offset", None)
            if choice.get("logprobs", None):
                del choice["logprobs"]
            if choice.get("index", None):
                del choice["index"]

            # # If the first token is a newline, remove it
            # if len(choice["text"]) == 0:
            #     continue
            # while choice["text"][0] == "\n":
            #     choice["text"] = choice["text"][1:]
            #     # And drop the first token logprob
            #     if choice.get("token_logprobs", None):
            #         choice["token_logprobs"] = choice["token_logprobs"][1:]
            #         # And update the offsets
            #         choice["logprob_offset"] = [offset - 1 for offset in choice["logprob_offset"]]
            #         choice["logprob_offset"] = choice["logprob_offset"][1:]

        return response

    def _logprobs(self, response):
        response = self._clean_json(response)
        prompt_offset = len(response["params"]["prompt"])
        if response["choices"][0].get("token_logprobs", None) is None:
            return response

        for c in response["choices"]:
            colorized = self._colorize(
                c["text"], c["token_logprobs"], [offset - prompt_offset for offset in c["logprob_offset"]]
            )
            c["text"] = colorized

        return response

    def _colorize(self, text, token_logprobs, offset):
        colorized_string = ""
        for i, logprob in enumerate(token_logprobs):
            start = offset[i]
            end = offset[i + 1] if i < len(offset) - 1 else len(text)
            # Start with a set of colors
            fg_, bg_ = ProbColors.choose_color(logprob)
            colorized_string += f"{bg(bg_)}{fg(fg_)}{text[start:end]}{attr(0)}"
        return colorized_string

    def _base(self, response):
        if self.format == "json":
            response_dict = {"choices": [{"text": c} for c in response]}
            return json.dumps(response_dict, indent=4)
        return "\n".join(response)


@dataclass
class BaseModel:
    operator: str = ""
    completion_url: str = ""
    config_base: Path = config_dir
    backup_path: Path = Path(config.get("backup_path", "/tmp/ttt/"))
    params: dict = field(default_factory=dict)
    format: str = config.get("format", "clean")
    echo_prompt: bool = config.get("echo_prompt", False)

    def __post_init__(self):
        self.backup_path.mkdir(parents=True, exist_ok=True)
        self.n = self.params.get("n", 1)

        self.formatter = Formatter(operator=self.operator, format=self.format, echo_prompt=self.echo_prompt)

    @staticmethod
    def token_position(token, text_offset):
        return {"start": text_offset, "end": text_offset + len(token)}

    def gen(self, prompt):
        return self.formatter.format_response([prompt] * self.n)


@dataclass
class OpenAIModel(BaseModel):
    operator: str = "OpenAI"
    completion_url: str = "https://api.openai.com/v1/completions"

    path: Path = config_dir / "openai.yaml"
    format: Optional[str] = None

    _list: list = field(default_factory=list)
    _params: dict = field(default_factory=dict)
    _config: dict = field(default_factory=dict)

    chat_models: list = field(default_factory=lambda: ["gpt-3.5-turbo-0301", "gpt-3.5-turbo"])
    large_models: list = field(
        default_factory=lambda: [
            "gpt-3.5-turbo",
            "gpt-3.5-turbo-0301",
            "text-davinci-003",
            "text-davinci-002",
            "code-davinci-002",
        ]
    )

    def __post_init__(self):
        self._config = yaml.load(self.path.read_text(), Loader=yaml.FullLoader)

        self._params = self._config.get("engine_params", {})
        for k, _ in self._params.items():
            if k in self.params:
                self._params[k] = self.params[k]

        self._list = self._config.get("models", [])

        openai.api_key = self._config.get("api_key", "")

        super().__post_init__()

    @property
    def list(self):
        return self._list

    def update_list(self):
        full_list = openai.Model.list()
        self._list = [m["id"] for m in full_list["data"]]
        self._config["models"] = self._list

    def save(self):
        self._config["engine_params"] = self._params
        self.path.write_text(yaml.dump(self._config))

    def gen(self, prompt):
        prompt_tokens = encoding.encode(prompt)
        max_tokens = 4000 if self._params["model"] in OpenAIModel().large_models else 2048
        if len(prompt_tokens) + int(self._params["max_tokens"]) > max_tokens:
            self._params["max_tokens"] = max_tokens - len(prompt_tokens)
            click.echo(
                "Prompt is too long. " f"Max tokens adjusted: {max_tokens}. " f"Prompt tokens: {len(prompt_tokens)}",
                err=True,
            )

        if self._params["model"] in self.chat_models:
            return self.formatter.format_response(self._chat(prompt))
        return self.formatter.format_response(self._gen(prompt))

    def _gen(self, prompt):
        self._params["prompt"] = prompt

        response = openai.Completion.create(**self._params).to_dict_recursive()
        self.backup(response)

        return response

    def _chat(self, prompt):
        params = self._params
        params["messages"] = [{"role": "user", "content": prompt}]
        if "prompt" in params:
            del params["prompt"]
        if "logprobs" in params:
            del params["logprobs"]

        response = openai.ChatCompletion.create(**params).to_dict_recursive()
        self.backup(response)

        for c in response["choices"]:
            c["text"] = c["message"]["content"]
            del c["message"]
        response["params"]["prompt"] = response["params"]["messages"][0]["content"]
        del response["params"]["messages"]

        return response

    def backup(self, response):
        timestamp = datetime.datetime.now().replace(microsecond=0).isoformat()
        response["params"] = self._params

        backup = self.backup_path / f"{timestamp}.json"
        backup.write_text(json.dumps(response, indent=4))

    @staticmethod
    def create_config(api_key):
        path = config_dir / "openai.yaml"
        from ttt.config import OPENAI_DEFAULT_PARAMS

        config = {
            "engine_params": OPENAI_DEFAULT_PARAMS,
            "api_key": api_key,
            "backup_path": str(BaseModel().backup_path),
            "models": [],
        }
        path.write_text(yaml.dump(config))

        model = OpenAIModel()
        model.update_list()
        model.save()


@dataclass
class GooseAIModel(OpenAIModel):
    operator: str = "GooseAI"
    completion_url: str = "https://api.goose.ai/v1/completions"


@dataclass
class AI21Model(BaseModel):
    operator: str = "AI21"
    completion_url: str = "https://api.ai21.com/studio/v1/completions"

    # def _gen(prompt):
    #     response = requests.post(
    #         f"https://api.ai21.com/studio/v1/{engine}/complete",
    #         headers={"Authorization": f"Bearer {api_key}"},
    #         json=request_json,
    #     )
